#!/usr/local/bin/perl

package IPAnalyze::IPmath;
use strict;

sub new {
	my $me = shift;
	return bless({}, $me);
}

sub cidr_check {
	my $self = shift;
	my $cidr = shift || die "cidr_check syntax error\n";
	my %cidrs = (32 => 1, 31 => 2, 30 => 4, 29 => 8,
	             28 => 16, 27 => 32, 26 => 64, 25 => 128,
	             24 => 256, 23 => 512, 22 => 1024,
	             21 => 2048, 20 => 4096, 19 => 8192,
	             18 => 16384, 17 => 32768, 16 => 65536,
	             15 => 131072, 14 => 262144, 13 => 524288,
	             12 => 1048576, 11 => 2097152, 10 => 4194304,
	             9 => 8388608, 8 => 16777216);

	return($cidrs{$cidr});
}

sub get_netblock {
	my $self = shift;
	my $netblock = shift || die "get_netblock syntax error1\n";
	my $cidr = shift || die "get_netblock syntax error2\n";

	my $u32 = unpack 'N', pack 'CCCC', split /\./, $netblock;
	my $bits = '1' x $cidr . '0' x (32 - $cidr);
	my $msk = unpack 'N', pack 'B*', $bits;
	my $net = join '.', unpack 'CCCC', pack 'N', $u32 & $msk;

	return($net);
}

sub get_broadcast {
	my $self = shift;
	my $netblock = shift || die "get_broadcast syntax error1\n";
	my $cidr = shift || die "get_broadcast syntax error2\n";

	my $u32 = unpack 'N', pack 'CCCC', split /\./, $netblock;
	my $bits = '1' x $cidr . '0' x (32 - $cidr);
	my $msk = unpack 'N', pack 'B*', $bits;
	my $broadcast = join '.', unpack 'CCCC', pack 'N', $u32 | ~$msk;

	return($broadcast);
}

sub get_ips {
	# QQQ no longer used in ipanalyze, but used elsewhere, like openrelay
	my $self = shift;
	my $netblock = shift || die "getips error1";
	my $cidr = shift || die "get_ips error2";

	if($cidr < 16){
		warn "will not compute for a cidr that big\n";
		return();
	}
	my $num = $self->cidr_check($cidr);

	if($num){
		my @ips;
		my ($oct1, $oct2, $oct3, $oct4) = split (/\./,$netblock);
		if($oct4 % $num){
			# bad network
			warn "get_ips: bad network: ${netblock}/${cidr}\n";
			return;
		}
		my ($a,$b,$c,$d) = ($oct1,$oct2,$oct3,$oct4); # copy
		while($num >= 256){
			while($d < 256){
				push @ips, "$oct1.$oct2.$c.$d";
				$d++;
			}
			$d=0;
			$num -= 256;
			$c++;
		}
		my $n = 0;
		while($n < $num){
			push @ips, "$oct1.$oct2.$oct3.$oct4";
			$oct4++;
			$n++;
		}

		return @ips;
	}else{
		warn "bad cidr: $cidr\n";
		return();
	}
}

sub get_netblocks {
	my $self = shift;
	my $netblock = shift || die "get_netblocks syntax error1\n";
	my $cidr = shift || die "get_netblocks syntax error2\n";
	my $for_cidr = shift || die "get_netblocks syntax error3\n";

	# to get all the /$for_cidr netblocks  between
	# $netblock and $broadcast 

	return() if($for_cidr < $cidr);

	my @netblocks;
	if($for_cidr == $cidr){
		# quick logic, if we want a /x out of a parent /x, just return it.
		push @netblocks, $netblock;
		return(@netblocks);
	}

	my $broadcast = $self->get_broadcast($netblock,$cidr);
	my $amount_in_for_cidr = $self->cidr_check($for_cidr);

	my($loct1,$loct2,$loct3,$loct4) = split(/\./, $netblock);
	my($hoct1,$hoct2,$hoct3,$hoct4) = split(/\./, $broadcast);

	my $num = $self->cidr_check($cidr);
	if($loct4 % $num){
		# bad network
		warn "get_netblocks: bad network: ${netblock}/${cidr}\n";
		return;
	}

	my ($oct1,$oct2,$oct3,$oct4) = ($loct1,$loct2,$loct3,$loct4); # a copy;
	while($loct2 <= $hoct2){
		while($loct3 <= $hoct3){
			while($loct4 <= $hoct4){
				push @netblocks, "$loct1.$loct2.$loct3.$loct4";
				$loct4 += $amount_in_for_cidr;

				if(($loct1 == $oct1 && $loct2 == $oct2 && $loct3 == $oct3 && $loct4 >= $hoct4) ||
				   ($loct4 >= 256)){
					if($for_cidr >= 24){
						$loct3++;
					}else{
						$loct3 += ($amount_in_for_cidr/256);
					}
					$loct4=0;
					last;
				}
			}
			if($loct3 == 256){
				$loct2++;
				$loct3=0;
				last;
			}
			last if($loct1 == $hoct1 && $loct2 == $hoct2 && $loct3 > $hoct3);
		}
		if($loct2 == 256){
			$loct1++;
			$loct2=0;
			last;
		}
		last if(($loct1 == $hoct1 && $loct2 == $hoct2 && $loct3 > $hoct3) ||
		        ($loct1 == $hoct1 && $loct2 > $hoct2) );
	}

	return(@netblocks);
}

sub get_network {
	my $self = shift;
	my $dbh = shift || die "get_network error1";
	my $cidr = shift || die "get_network error2";
	my $area = shift || die "get_network error3";

	my $DEBUG = 1; # 0 = off, 1 = minimal, 2 = verbose

	my $starttime;
	if($DEBUG){
		$starttime = time();
		warn "starting get_network at $starttime\n";
	}
	my $sql = 'SELECT id,netblock,cidr FROM ipanalyze.parent_networks WHERE area = ' . $dbh->quote($area);
	$sql .= ' AND allow_use = 1 ORDER BY netblock DESC';
	my $sth = $dbh->prepare($sql);
	$sth->execute;
	my %bigger; # use as last resort
	while(my $row=$sth->fetchrow_arrayref){
		my $id = $row->[0];
		my $parent_netblock = $row->[1];
		my $parent_cidr = $row->[2];
	
		# QQQ could set skip_Cache = 2 when %bigger and loop through it later
		my $sqlc = 'SELECT COUNT(*) FROM ipanalyze.skip_cache WHERE parent_id = ' . $dbh->quote($id);
		$sqlc .= ' AND cidr = ' . $dbh->quote($cidr);
		my $sthc = $dbh->prepare($sqlc);
		$sthc->execute;
		my $rowc = $sthc->fetchrow_arrayref;
		if($rowc->[0]){
			warn "skipping parent_id $id ($parent_netblock/$parent_cidr)\n" if($DEBUG > 1);
			next;
		}

		my $sqlb = 'SELECT netblock,cidr FROM data WHERE parent_id = ' . $dbh->quote($id);
		$sqlb .= ' ORDER BY netblock DESC';
		warn "sqlb: $sqlb\n" if($DEBUG > 1);
		my $sthb = $dbh->prepare($sqlb);
		$sthb->execute;
		my %children_networks;
		while(my $rowb = $sthb->fetchrow_arrayref){
			$children_networks{$rowb->[0]} = $rowb->[1];
			warn "adding children_networks{$rowb->[0]} = $rowb->[1]\n" if($DEBUG > 1);
		}

		my $can_make_bigger = 0;

		my @netblocks = $self->get_netblocks($parent_netblock,$parent_cidr,$cidr); # all the /28s ($cidr) in 10.0.0.0/20
		my $blocks_to_skip = 0;
		my %skip;
		foreach my $netblock (@netblocks){
		
			if($blocks_to_skip){
				warn "skipping[0] $netblock\n" if($DEBUG > 1);
				$blocks_to_skip--;
				next;
			}
			if(exists($skip{$netblock})){
				warn "skipping[1] $netblock\n" if($DEBUG > 1);
				delete $skip{$netblock}; # do we care about memory ?
				next;
			}

			# check to see if any netblocks (down to /30s) are used in parent_netblock
			my @possible = $self->get_netblocks($netblock,$cidr,30); # give me all the /30s in $network/$cidr
		
			my $used = 0;
			foreach my $ip (@possible){
				if(exists($children_networks{$ip})){
					$used = 1;
					last;
				}
			}

			if($used){
				next;
			}else{
				warn "* -> $netblock seems available\n" if($DEBUG > 1);
			}

			# check to see if this netblock is a child of another allocation
			
			my $bigger_cidr = ($cidr - 1);
			my $used_by_parent = 0;
			my $times = 1;
			while($bigger_cidr >= 19){  # 19 is the biggest possible child in ipanalyze QQQ
				my $superblock = $self->get_netblock($netblock,$bigger_cidr);
	
				if($children_networks{$superblock} == $bigger_cidr && ($superblock)){
					warn "${superblock}/${bigger_cidr} is used and encompasses $netblock/$cidr\n" if($DEBUG > 1);
					$used_by_parent = 1;
					# skip remaining questions until we're out of the $superblock/$bigger_cidr that we just discovered
					# cant do this the way we did below, we could be checking 196 and have 192/27 used - 
					foreach my $skipblock ($self->get_netblocks($superblock,$bigger_cidr,$cidr)){
						$skip{$skipblock} = 1;
					}
					last;
				}
				$bigger_cidr -= 1;
				$times++;
			}
			if($used_by_parent){
				warn "$netblock used by parent\n" if($DEBUG > 1);
			}
			unless($used_by_parent){
				# check to see if this netblock could make a bigger cidr than what we're asking for
				my $bigger_cidr = ($cidr - 1);
				while($bigger_cidr >= 19){
					my $superblock = $self->get_netblock($netblock,$bigger_cidr);
					my @middleblocks = $self->get_netblocks($superblock,$bigger_cidr,30);

					my $partially_used;
					foreach my $block (@middleblocks){
						if(exists($children_networks{$block})){
							$partially_used = 1;
							last;
						}
					}
					if($partially_used){
						my $smaller_cidr = ($bigger_cidr + 1);
						if($bigger{$smaller_cidr}{$netblock}){
							warn "$netblock can't quite make the /$bigger_cidr\n" if($DEBUG > 1);
							last;
						}else{
							warn "$netblock can't make a /$bigger_cidr - FINISHED\n" if($DEBUG);
							if($DEBUG){
								my $delta = (time() - $starttime);
								warn "network allocated in $delta seconds\n";
							}
							return($netblock,$id);
						}

					}else{
						# QQQ if bigger{cidr - 1} exists , dont go further than that -
						# ie, if we're looking for a /30, and we see something can make a /29,
						# we dont need to check for a /28.
						# same as if we know of a bigger{/28}, dont check for a /27 or /26 ...
						# but then what about the skipping code below.
						warn "See $netblock is available, but it can make $netblock/$bigger_cidr\n" if($DEBUG > 1);
						$can_make_bigger = 1;
						my $smaller_cidr = ($bigger_cidr + 1);
						delete $bigger{$smaller_cidr}{$netblock}; # if it exists, or not.
						$bigger{$bigger_cidr}{$netblock} = $id; # available, but use as a last resort
					}
					$bigger_cidr -= 1;
				}
				# skip remaining questions until we're out of the $netblock/$bigger_cidr that we just discovered
				# how many _more_ /$cidr's are in $bigger_cidr ?
				my $ips = $self->cidr_check($cidr);
				my $smaller_cidr = ($bigger_cidr + 1);
				my $totalips = $self->cidr_check($smaller_cidr);
				$blocks_to_skip = (($totalips / $ips) - 1); # minus 1 for the one we just did
				warn "blocks_to_skip[0] = $blocks_to_skip\n" if($DEBUG > 1);

				warn "will skip $blocks_to_skip netblocks ((totalips: $totalips / ips: $ips) - 1)\n" if($DEBUG > 1);
			}
		}
		unless($can_make_bigger){
			warn "CACHING: no available /$cidr networks in parent_id $id ($parent_netblock/$parent_cidr)\n" if($DEBUG);
			my $sqld = 'INSERT INTO skip_cache (parent_id,cidr,timestamp) VALUES (';
			$sqld .= $dbh->quote($id) . ',' . $dbh->quote($cidr) . ',' . $dbh->quote(time()) . ')';
			my $sthd = $dbh->prepare($sqld);
			$sthd->execute;
		}
	}
	foreach my $size (reverse sort keys %bigger){
		foreach my $netblock (keys %{$bigger{$size}}){
			warn "breaking $netblock/$size for a /$cidr\n";
			if($DEBUG){
				my $delta = (time() - $starttime);
				warn "network allocated in $delta seconds\n";
			}
			return($netblock,$bigger{$size}{$netblock});
		}
	}
}

return 1;
